#ifndef lint
static const char rcsid[] = "$Header: /net/hopkins/home/tristan/mpeg/dec/RCS/pref.c,v 1.17 1999/09/08 12:54:10 tristan Exp tristan $";
static const char copyright[] = "Copyright (c) 1997, MpegTV LLC";
#endif

#include "mtvp/misc.h"

#include "mtvp/msg.h"

#include "pref.h"

/* #define DEBUG_PREF */

extern MTVP_MSG_STRUCT Msg;

/*--------------------------------------------------------------------------+
| Macros																	|
+--------------------------------------------------------------------------*/
#define MTVP_MSG_PTR (&Msg)

/*--------------------------------------------------------------------------+
| Global variables															|
+--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------+
| Static variables															|
+--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------+
| enter_key																	|
+--------------------------------------------------------------------------*/
static void
enter_key(const ENTRY *p)
{
	/* enter key in hash table */
	
	if (hsearch(*p, ENTER) == NULL) {
		int	status = errno;
		MTVP_FATAL_ERROR(("hsearch: cannot ENTER key = \"%s\"\n", p->key));
	}
}

	
#ifdef DEBUG_PREF
/*--------------------------------------------------------------------------+
| dump_keys																	|
+--------------------------------------------------------------------------*/
static void
dump_keys(
	const ENTRY *keys,
	u_int		keys_count
)
{
	u_int	i;

	for (i = 0; i < keys_count; i++) {
		ENTRY	entry;
		const ENTRY	*p;

		entry.key = keys[i].key;
		
		if ((p = hsearch(entry, FIND)) == NULL || p->data == NULL) {
			abort_now();
		}
		
		printf("\"%s\" = \"%s\"\n", p->key, p->data);
	}
#if 0 /* Tristan Savatier - Tue Oct 14 12:43:29 PDT 1997 */
//	printf("pref: end of dump\n");
#endif /* 0 - Tristan Savatier - Tue Oct 14 12:43:29 PDT 1997 */
}
#endif /* DEBUG_PREF */
	

/*--------------------------------------------------------------------------+
| pref_init_keys															|
+--------------------------------------------------------------------------*/
void
pref_init_keys(
	const ENTRY *keys,
	u_int		keys_count
)
{
	u_int	i;

	/* create hash table */
	if (hcreate(keys_count) == 0) {
		int	status = errno;
		MTVP_FATAL_ERROR(("hcreate"));
	}
	
	/* enter keys */
	for (i = 0; i < keys_count; i++) {
		MTVP_MESSAGE(("pref: entering key %s = %s", keys[i].key, keys[i].data));
		enter_key(&keys[i]);
	}
	
#ifdef DEBUG_PREF
	dump_keys(keys, keys_count);
#endif /* DEBUG_PREF */
}

/*--------------------------------------------------------------------------+
| read_prop																	|
+--------------------------------------------------------------------------*/
static int
read_prop(file)
	FILE	*file;
{
	char	key[1024];
	char	value[1024];
	int		n;
	
	/*
	 * This routine reads the preference file and set the
	 * values (entry.data) for the first recognized key.
	 * If no key could be found before EOF, it
	 * returns -1. Otherwise it returns 0.
	 * This routine must be called until it returns -1 in order
	 * to parse the entire file.
	 * A key is recognised if it is already present in the hash table.
	 * The value is a string (without any leading and trailing spaces).
	 * The value is stored in a memory_alloc buffer.
	 * Keys that are not present in the hash table are ignored.
	 * The syntax for KEY = "value string" is a line of the form
	 * KEY: value string
	 * Comments start with # and extend until the end of the line.
	 * Empty lines and comments are ignored.
	 * Anything that is not recognized is ignored too.
     */
     
	for (; ; ) {
	
		/* skip every space or #-starting comments */
		for (; ; ) {
			int		c;
			
			/* skip everything until non-space */
			for (; ; ) {
				c = getc(file);
				if (c == EOF || !isspace(c)) {
					break;
				}
			}
	
			if (c == EOF) {
				break;
			}
			if (c == '#') {
				/* skip everything after # until end-of-line */
				for (; ; ) {
					c = getc(file);
					if (c == EOF || c == '\n') {
						break;
					}
				}
				if (c == EOF) {
					break;
				}
			} else {
				/* restore character and exit this loop */
				ungetc(c, file);
				break;
			}
		}
		
		key[sizeof key - 1] = 0;
		value[sizeof value - 1] = 0;
				
		n = fscanf(file, " %[^ 	:]:%*[ ]%[^#\n]\n", key, value);
		if (n == 1) {
			value[0] = 0;
		} else if (n != 2) {
		
#if 0 /* Tristan Savatier - Fri Oct 10 17:32:48 PDT 1997 */
//			printf("n = %d\n", n);
//			if (n == 1) {
//				printf("key = \"%s\"\n", key);
//			}
#endif /* 0 - Tristan Savatier - Fri Oct 10 17:32:48 PDT 1997 */
			fscanf(file, "%[^\n]\n", value);
#if 0 /* Tristan Savatier - Fri Oct 10 17:32:53 PDT 1997 */
//			printf("crap = \"%s\"\n", value);
#endif /* 0 - Tristan Savatier - Fri Oct 10 17:32:53 PDT 1997 */
			if (feof(file)) {
				return -1;
			}
			continue;
		}
		if (key[sizeof key - 1] != 0 || value[sizeof value - 1] != 0) abort_now();
	
		/* remove any trailing spaces from value */
		{
			char *p = value + strlen(value) - 1;
			while (isspace(*p) && p >= value) p--;
			*(p + 1) = 0;
		}
		
#ifdef DEBUG_PREF
		printf("key = \"%s\"\n", key);
		printf("value = \"%s\"\n", value);
#endif /* DEBUG_PREF */
		MTVP_MESSAGE(("pref: adding key = \"%s\", value = \"%s\"", key, value));
	
		{
			ENTRY	entry;
			ENTRY	*p;

			entry.key = key;
	
			/* look for key in hash table */
			p = hsearch(entry, FIND);			
			
			if (p == NULL) {
				MTVP_MESSAGE(("pref: key unknown (ignored): \"%s\"", key));
				continue;
			}
	

#ifndef HAVE_SEARCH_H
			/*
			 * When HAVE_SEARCH_H is not defined, we use an emulation of hsearch etc
			 * based on the DB interface (see search.c).
			 *
			 * With this emulation, changing the value of an existing key
			 * requires calling hsearch(entry, ENTER) again, with the new value.
			 */
			 
			/* memory_alloc buffer for value */
			entry.key = key;
			entry.data = strdup(value);
			if (entry.data == NULL) {
				int status = errno;
				MTVP_FATAL_ERROR(("cannot allocate memory"));
			}
			if ((p = hsearch(entry, ENTER)) == NULL) {
				int status = errno;
				MTVP_FATAL_ERROR(("hsearch: cannot ENTER key = \"%s\"\n", entry.key));
			}
			if (p->data == NULL) {
				MTVP_FATAL(("hsearch: p->data == NULL after ENTER\n"));
			}
			if (strcmp(p->data, value) != 0) {
				MTVP_FATAL(("hsearch: strcmp(p->data, value) != 0 after ENTER\n"));
			}
#else
			/*
			 * Patching the value of an existing key seems to work, although undocumented.
			 */

			/* memory_alloc buffer for value */
			p->data = strdup(value);
			if (p->data == NULL) {
				int status = errno;
				MTVP_FATAL_ERROR(("cannot allocate memory"));
			}
#endif
		}
	
		/* return after one key has been found */
		return 0;
	}
	/* should never reach that! */
	abort_now();
	return 0;
}

/*--------------------------------------------------------------------------+
| pref_get_value															|
+--------------------------------------------------------------------------*/
const char *
pref_get_value(const char *key) 
{
	ENTRY	entry;
	const ENTRY	*p;
	
	/* return the value of a key that exists in the hash table */
	
	entry.key = (char *)key;
	
	p = hsearch(entry, FIND);
	
	if (p == NULL) {
		MTVP_MESSAGE(("pref: unknown key: %s", key));
		return NULL;
	}
		
	if (p->data == NULL) {
		MTVP_MESSAGE(("pref: %s undefined", key));
		return NULL;
	}

	MTVP_MESSAGE(("pref: key = \"%s\" has value \"%s\"", key, p->data));
		
	return p->data;
}

/*--------------------------------------------------------------------------+
| pref_get_value_bool														|
+--------------------------------------------------------------------------*/
int
pref_get_value_bool(
	const char *key,
	Bool		*p
)
{
	const char *value;
	
	value = pref_get_value(key);
	
	if (value == NULL) {
		return -1;
	}

	if (strings_compare_no_case(value, "True") == 0) {
		*p = True;
		return 0;
	} else if (strings_compare_no_case(value, "False") == 0) {
		*p = False;
		return 0;
	} else {
		MTVP_MESSAGE(("pref: %s = \"%s\" should be True or False (value ignored)", key, value));
		return -1;
	}
}

/*--------------------------------------------------------------------------+
| pref_get_value_int														|
+--------------------------------------------------------------------------*/
int
pref_get_value_int(
	const char *key,
	int		*p,
	int		min,
	int		max
)
{
	const char *value;
	int		n;
	
	value = pref_get_value(key);

	if (value == NULL) {
		return -1;
	}

	n = atoi(value);
	
	if (min <= max) {
		if (n < min || n > max) {
			MTVP_MESSAGE(("pref: %s = %d out of range [%d, %d] (value ignored)", key, n, min, max));
			return -1;
		}
	}
	*p = n;
	return 0;
}

/*--------------------------------------------------------------------------+
| pref_get_value_double														|
+--------------------------------------------------------------------------*/
int
pref_get_value_double(key, p, min, max)
	const char *key;
	double		*p;
	double		min, max;
{
	const char *value;
	double		n;
	char *q;
	
	value = pref_get_value(key);

	if (value == NULL) {
		return -1;
	}

	n = strtod(value, &q);
	if (q == value) {
		MTVP_MESSAGE(("pref: %s = \"%s\" not a double (value ignored)", key, value));
		return -1;
	}
	
	if (min <= max && n < min || n > max) {
		MTVP_MESSAGE(("pref: %s = %6.4f out of range [%6.4f, %6.4f] (value ignored)", key, n, min, max));
		return -1;
	}
	*p = n;
	return 0;
}

#if 0 /* Tristan Savatier - Tue Oct 14 01:38:51 PDT 1997 */
///*--------------------------------------------------------------------------+
//| pref_get_value_double														|
//+--------------------------------------------------------------------------*/
//int
//pref_get_value_double(key, p, min, max)
//	const char *key;
//	double		*p;
//	double		min, max;
//{
//	const char *value;
//	double		n;
//		
//	if (sscanf(value, "%lf", &n) != 1) {
//		MTVP_MESSAGE(("pref: %s = \"%s\" not a double (value ignored)", key, value));
//		return -1;
//	}
//	
//	if (min <= max && n < min || n > max) {
//		MTVP_MESSAGE(("pref: %s = %6.4f out of range [%6.4f, %6.4f] (value ignored)", key, n, min, max));
//		return -1;
//	}
//	*p = n;
//	return 0;
//}
#endif /* 0 - Tristan Savatier - Tue Oct 14 01:38:51 PDT 1997 */


/*--------------------------------------------------------------------------+
| pref_load																	|
+--------------------------------------------------------------------------*/
int
pref_load(name)
	const char *name;
{
	FILE	*file;
	
	/* read key values from file */

	MTVP_MESSAGE(("pref: reading from file %s", name));
	
	file = fopen(name, "r");
	
	if (file == NULL) {
		int	status = errno; if (status == 0) status = -1;
		MTVP_MESSAGE_ERROR(("pref: cannot open %s", name));
		return status;
	}
	
	/* parse file until no more keys are found */
	while (read_prop(file) == 0);

	fclose(file);

	return 0;
}

/* eof */
